@@ -1,20 +1,21 @@
 import logging
-from typing import Dict, Tuple, List, Any, Union
+import warnings
+from typing import Any, Dict, List, Tuple, Union
 
 import numpy
-from overrides import overrides
 import torch
-from torch.nn.modules.linear import Linear
-from torch.nn.modules.rnn import LSTMCell
-
-from allennlp.common.util import START_SYMBOL, END_SYMBOL
+from allennlp.common.lazy import Lazy
+from allennlp.common.util import END_SYMBOL, START_SYMBOL
 from allennlp.data import TextFieldTensors, Vocabulary
 from allennlp.models.model import Model
-from allennlp.modules import Attention, TextFieldEmbedder, Seq2SeqEncoder
+from allennlp.modules import Attention, Seq2SeqEncoder, TextFieldEmbedder
 from allennlp.modules.token_embedders import Embedding
 from allennlp.nn import InitializerApplicator, util
-from allennlp.training.metrics import Metric, BLEU
 from allennlp.nn.beam_search import BeamSearch
+from allennlp.training.metrics import BLEU, Metric
+from overrides import overrides
+from torch.nn.modules.linear import Linear
+from torch.nn.modules.rnn import LSTMCell
 
 logger = logging.getLogger(__name__)
 
@@ -22,7 +23,7 @@ logger = logging.getLogger(__name__)
 @Model.register("copynet_seq2seq")
 class CopyNetSeq2Seq(Model):
     """
-    This is an implementation of [CopyNet](https://arxiv.org/pdf/1603.06393).
+    This is an implementation of [CopyNet](https://api.semanticscholar.org/CorpusID:8174613).
     CopyNet is a sequence-to-sequence encoder-decoder model with a copying mechanism
     that can copy tokens from the source sentence into the target sentence instead of
     generating all target tokens only from the target vocabulary.
@@ -46,10 +47,13 @@ class CopyNetSeq2Seq(Model):
     attention : `Attention`, required
         This is used to get a dynamic summary of encoder outputs at each timestep
         when producing the "generation" scores for the target vocab.
-    beam_size : `int`, required
-        Beam width to use for beam search prediction.
-    max_decoding_steps : `int`, required
-        Maximum sequence length of target predictions.
+    label_smoothing : `float`, optional (default = `None`)
+        Whether or not to apply label smoothing to the cross-entropy loss.
+        For example, with a label smoothing value of 0.2, a 4 class classification
+        target would look like `[0.05, 0.05, 0.85, 0.05]` if the 3rd class was
+        the correct label.
+    beam_search : `BeamSearch`, optional (default = `Lazy(BeamSearch)`)
+        This is used to during inference to select the tokens of the decoded output sequence.
     target_embedding_dim : `int`, optional (default = `30`)
         The size of the embeddings for the target vocabulary.
     copy_token : `str`, optional (default = `'@COPY@'`)
@@ -76,14 +80,15 @@ class CopyNetSeq2Seq(Model):
         source_embedder: TextFieldEmbedder,
         encoder: Seq2SeqEncoder,
         attention: Attention,
-        beam_size: int,
-        max_decoding_steps: int,
+        label_smoothing: float = None,
+        beam_search: Lazy[BeamSearch] = Lazy(BeamSearch),
         target_embedding_dim: int = 30,
         copy_token: str = "@COPY@",
         target_namespace: str = "target_tokens",
         tensor_based_metric: Metric = None,
         token_based_metric: Metric = None,
         initializer: InitializerApplicator = InitializerApplicator(),
+        **kwargs
     ) -> None:
         super().__init__(vocab)
         self._target_namespace = target_namespace
@@ -124,6 +129,7 @@ class CopyNetSeq2Seq(Model):
             num_embeddings=self._target_vocab_size, embedding_dim=target_embedding_dim
         )
         self._attention = attention
+        self._label_smoothing = label_smoothing
         self._input_projection_layer = Linear(
             target_embedding_dim + self.encoder_output_dim * 2, self.decoder_input_dim
         )
@@ -142,8 +148,21 @@ class CopyNetSeq2Seq(Model):
         self._output_copying_layer = Linear(self.encoder_output_dim, self.decoder_output_dim)
 
         # At prediction time, we'll use a beam search to find the best target sequence.
-        self._beam_search = BeamSearch(
-            self._end_index, max_steps=max_decoding_steps, beam_size=beam_size
+        # For backwards compatibility, check if beam_size or max_decoding_steps were passed in as
+        # kwargs. If so, update the BeamSearch object before constructing and raise a DeprecationWarning
+        deprecation_warning = (
+            "The parameter {} has been deprecated."
+            " Provide this parameter as argument to beam_search instead."
+        )
+        beam_search_extras = {}
+        if "beam_size" in kwargs:
+            beam_search_extras["beam_size"] = kwargs["beam_size"]
+            warnings.warn(deprecation_warning.format("beam_size"), DeprecationWarning)
+        if "max_decoding_steps" in kwargs:
+            beam_search_extras["max_steps"] = kwargs["max_decoding_steps"]
+            warnings.warn(deprecation_warning.format("max_decoding_steps"), DeprecationWarning)
+        self._beam_search = beam_search.construct(
+            end_index=self._end_index, vocab=self.vocab, **beam_search_extras
         )
 
         initializer(self)
@@ -157,6 +176,7 @@ class CopyNetSeq2Seq(Model):
         metadata: List[Dict[str, Any]],
         target_tokens: TextFieldTensors = None,
         target_token_ids: torch.Tensor = None,
+        weight: torch.Tensor = None,
     ) -> Dict[str, torch.Tensor]:
         """
         Make foward pass with decoder logic for producing the entire target sequence.
@@ -183,6 +203,13 @@ class CopyNetSeq2Seq(Model):
         target_token_ids : `torch.Tensor`, optional (default = `None`)
             A tensor of shape `(batch_size, target_sequence_length)` which indicates which
             tokens in the target sequence match tokens in the source sequence.
+        weight : `torch.Tensor`, optional (default = `None`)
+            A optional tensor of shape `(batch_size,)` or `(batch_size, 1)` which determines
+            how to weight each instance in the batch when calculating the loss.
+            The default of `None` is equivalent to `weights = torch.tensor([1.0] * batch_size)`.
+            The final loss is calculated as `-(weight * log_likelihood).sum() / batch_size`,
+            where `log_likelihood` is a tensor of shape `(batch_size,)`, representing the overall
+            log likelihood of the gold target sequence.
 
         # Returns
 
@@ -194,7 +221,7 @@ class CopyNetSeq2Seq(Model):
 
         if target_tokens:
             state = self._init_decoder_state(state)
-            output_dict = self._forward_loss(target_tokens, target_token_ids, state)
+            output_dict = self._forward_loss(target_tokens, target_token_ids, state, weight=weight)
         else:
             output_dict = {}
 
@@ -425,8 +452,23 @@ class CopyNetSeq2Seq(Model):
         gen_mask = (target_tokens != self._oov_index) | (target_to_source.sum(-1) == 0)
         log_gen_mask = (gen_mask + util.tiny_value_of_dtype(log_probs.dtype)).log().unsqueeze(-1)
         # Now we get the generation score for the gold target token.
-        # shape: (batch_size, 1)
-        generation_log_probs = log_probs.gather(1, target_tokens.unsqueeze(1)) + log_gen_mask
+        if self._label_smoothing is not None and self._label_smoothing > 0.0:
+            smoothing_value = self._label_smoothing / target_size
+            # Create the smoothed targets to be multiplied with `log_probs`
+            # shape: (batch_size, target_vocab_size + source_sequence_length)
+            smoothed_targets = torch.full_like(log_probs, smoothing_value).scatter_(
+                1, target_tokens.unsqueeze(1), 1.0 - self._label_smoothing + smoothing_value
+            )
+            generation_log_probs = log_probs * smoothed_targets
+            # shape: (batch_size, 1)
+            generation_log_probs = generation_log_probs.sum(-1, keepdim=True)
+        else:
+            # Contribution to the negative log likelihood only comes from the exact indices
+            # of the targets, as the target distributions are one-hot. Here we use torch.gather
+            # to extract the indices which contribute to the loss.
+            # shape: (batch_size, 1)
+            generation_log_probs = log_probs.gather(1, target_tokens.unsqueeze(1))
+        generation_log_probs = generation_log_probs + log_gen_mask
         # ... and add the copy score to get the step log likelihood.
         # shape: (batch_size, 1 + source_sequence_length)
         combined_gen_and_copy = torch.cat((generation_log_probs, copy_log_probs), dim=-1)
@@ -440,6 +482,7 @@ class CopyNetSeq2Seq(Model):
         target_tokens: TextFieldTensors,
         target_token_ids: torch.Tensor,
         state: Dict[str, torch.Tensor],
+        weight: torch.Tensor = None,
     ) -> Dict[str, torch.Tensor]:
         """
         Calculate the loss against gold targets.
@@ -527,7 +570,13 @@ class CopyNetSeq2Seq(Model):
         # shape: (batch_size,)
         log_likelihood = (log_likelihoods * target_mask).sum(dim=-1)
         # The loss is the negative log-likelihood, averaged over the batch.
-        loss = -log_likelihood.sum() / batch_size
+        if weight is None:
+            loss = -log_likelihood.sum() / batch_size
+        else:
+            # shape: (batch_size,)
+            if len(weight.shape) > 1:
+                weight = weight.squeeze()
+            loss = -(weight * log_likelihood).sum() / batch_size
 
         return {"loss": loss}
 
@@ -795,9 +844,9 @@ class CopyNetSeq2Seq(Model):
         step : `int`
             The time step in beam search decoding.
 
-        Notes
-        -----
-        `group_size` != `batch_size`. In fact, `group_size` = `batch_size * beam_size`.
+        # Notes
+
+        `group_size != batch_size`. In fact, `group_size = batch_size * beam_size`.
         """
         _, source_sequence_length = state["source_to_target"].size()
 
