@@ -1,5 +1,7 @@
+from allennlp_models.vision.models.nlvr2 import Nlvr2Model
 from allennlp_models.vision.models.vision_text_model import VisionTextModel
 from allennlp_models.vision.models.visual_entailment import VisualEntailmentModel
+from allennlp_models.vision.models.vilbert_image_retrieval import ImageRetrievalVilbert
 from allennlp_models.vision.models.vilbert_vqa import VqaVilbert
 from allennlp_models.vision.models.heads.vqa_head import VqaHead
 from allennlp_models.vision.models.heads.visual_entailment_head import VisualEntailmentHead
